from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field, AfterValidator
from pydantic.types import StringConstraints
from publicsuffix2 import PublicSuffixList


class Regions(str, Enum):
    ARGENTINA = 'ar-es'
    AUSTRALIA = 'au-en'
    AUSTRIA = 'at-de'
    BELGIUM_FR = 'be-fr'
    BELGIUM_NL = 'be-nl'
    BRAZIL = 'br-pt'
    BULGARIA = 'bg-bg'
    CANADA_EN = 'ca-en'
    CANADA_FR = 'ca-fr'
    CATALONIA = 'ct-ca'
    CHILE = 'cl-es'
    CHINA = 'cn-zh'
    COLOMBIA = 'co-es'
    CROATIA = 'hr-hr'
    CZECH_REPUBLIC = 'cz-cs'
    DENMARK = 'dk-da'
    ESTONIA = 'ee-et'
    FINLAND = 'fi-fi'
    FRANCE = 'fr-fr'
    GERMANY = 'de-de'
    GREECE = 'gr-el'
    HONG_KONG = 'hk-tzh'
    HUNGARY = 'hu-hu'
    ICELAND = 'is-is'
    INDIA_EN = 'in-en'
    INDONESIA_EN = 'id-en'
    IRELAND = 'ie-en'
    ISRAEL_EN = 'il-en'
    ITALY = 'it-it'
    JAPAN = 'jp-jp'
    KOREA = 'kr-kr'
    LATVIA = 'lv-lv'
    LITHUANIA = 'lt-lt'
    MALAYSIA_EN = 'my-en'
    MEXICO = 'mx-es'
    NETHERLANDS = 'nl-nl'
    NEW_ZEALAND = 'nz-en'
    NORWAY = 'no-no'
    PAKISTAN_EN = 'pk-en'
    PERU = 'pe-es'
    PHILIPPINES_EN = 'ph-en'
    POLAND = 'pl-pl'
    PORTUGAL = 'pt-pt'
    ROMANIA = 'ro-ro'
    RUSSIA = 'ru-ru'
    SAUDI_ARABIA = 'xa-ar'
    SINGAPORE = 'sg-en'
    SLOVAKIA = 'sk-sk'
    SLOVENIA = 'sl-sl'
    SOUTH_AFRICA = 'za-en'
    SPAIN_CA = 'es-ca'
    SPAIN_ES = 'es-es'
    SWEDEN = 'se-sv'
    SWITZERLAND_DE = 'ch-de'
    SWITZERLAND_FR = 'ch-fr'
    TAIWAN = 'tw-tzh'
    THAILAND_EN = 'th-en'
    TURKEY = 'tr-tr'
    US_ENGLISH = 'us-en'
    US_SPANISH = 'us-es'
    UKRAINE = 'ua-uk'
    UNITED_KINGDOM = 'uk-en'
    VIETNAM_EN = 'vn-en'


class Timespans(str, Enum):
    PAST_DAY = 'd'
    PAST_WEEK = 'w'
    PAST_MONTH = 'm'
    PAST_YEAR = 'y'


FileType = Annotated[str, StringConstraints(pattern=r'^[a-zA-Z0-9]{2,10}$')]
Keyword = Annotated[str, StringConstraints(pattern=r'^[^\s]+$')]

psl = PublicSuffixList()


def to_list(val: str | list[str] | None) -> list[str]:
    if val is None:
        return []
    return val.split(' ') if isinstance(val, str) else val


def group_includes(values: list[str], op: str | None = None) -> str:
    if not values:
        return ''

    if len(values) == 1:
        return f'{op}:{values[0]}' if op else f'"{values[0]}"'

    if op:
        parts = [f'{op}:{v}' for v in values]
    else:
        parts = [f'"{v}"' for v in values]

    return f'({" | ".join(parts)})'


def group_excludes(op: str, values: list[str]) -> list[str]:
    return [f'-{op}:{v}' for v in values]


def validate_tld(v: str | list[str] | None):
    # TODO: make type a generic type
    if not v:
        return v

    tlds = v if isinstance(v, list) else [v]
    for tld in tlds:
        if not psl.get_tld(tld, strict=True):
            raise ValueError(f'{tld!r} is an invalid top-level domain according to https://publicsuffix.org')

    return v


class Filters(BaseModel):
    # fmt: off
    region: Regions | None = Field(None, description ='Only show results from specific regions')
    time_span: Timespans | None = Field(None, description='Timespan for the search')

    tlds: Annotated[str | list[str] | None, AfterValidator(validate_tld)] = Field(None, description='Only show results from specific top-level domains (e.g., .gov, .edu)')
    sites: str | list[str] | None = Field(None, description='Only show results from specific domains')
    filetype: FileType | None = Field(None, description='Only show documents that are a specific file type. Note: Google only supports one filetype per search.')
    https_only: bool = Field(False, description='Only show websites that support HTTPS')

    exclude_tlds: Annotated[str | list[str] | None, AfterValidator(validate_tld)] = Field(None, description='Exclude results from specific top-level domains')
    exclude_sites: str | list[str] | None = Field(None, description='Exclude results from specific domains')
    exclude_filetypes: FileType | list[FileType] | None = Field(None, description='Exclude documents with specific file types')
    exclude_https: bool = Field(False, description='Exclude HTTPS pages')

    any_keywords: Keyword | list[Keyword] | None = Field(None, description='Require at least one word anywhere in the page')
    all_keywords: Keyword | list[Keyword] | None = Field(None, description='Require specific words anywhere in the page')
    exact_phrases: str | list[str] | None = Field(None, description='Include results with exact phrases')

    exclude_all_keywords: Keyword | list[Keyword] | None = Field(None, description='Exclude pages containing certain words')
    exclude_exact_phrases: str | list[str] | None = Field(None, description='Exclude results with exact phrases')

    in_title: str | list[str] | None = Field(None, description='Require specific words in the title')
    in_url: str | list[str] | None = Field(None, description='Require specific words in the URL')
    in_text: str | list[str] | None = Field(None, description='Require specific words in the page text')

    not_in_title: str | list[str] | None = Field(None, description='Exclude pages with specific words in the title')
    not_in_url: str | list[str] | None = Field(None, description='Exclude pages with specific words in the URL')
    not_in_text: str | list[str] | None = Field(None, description='Exclude pages with specific words in the page text')
    # fmt: on

    def compile_filters(self) -> str:
        filters = []

        filters.append(group_includes(to_list(self.sites), 'site'))
        filters.append(group_includes(to_list(self.tlds), 'site'))

        filters.append(group_includes(to_list(self.filetype), 'filetype'))
        filters.append(group_includes(to_list(self.any_keywords)))
        filters.extend([f'"{w}"' for w in to_list(self.all_keywords)])

        if self.exact_phrases:
            phrase_list = self.exact_phrases if isinstance(self.exact_phrases, list) else [self.exact_phrases]
            filters.extend([f'"{phrase}"' for phrase in phrase_list])

        if self.https_only:
            filters.append('inurl:https')

        filters.extend([f'intitle:{w}' for w in to_list(self.in_title)])
        filters.extend([f'inurl:{w}' for w in to_list(self.in_url)])
        filters.extend([f'intext:{w}' for w in to_list(self.in_text)])

        # Negative Filters
        if self.exclude_exact_phrases:
            phrase_list = (
                self.exclude_exact_phrases
                if isinstance(self.exclude_exact_phrases, list)
                else [self.exclude_exact_phrases]
            )
            filters.extend([f'-"{phrase}"' for phrase in phrase_list])

        filters.extend(group_excludes('site', to_list(self.exclude_sites)))
        filters.extend(group_excludes('site', to_list(self.exclude_tlds)))

        if self.exclude_https:
            filters.append('-inurl:https')

        filters.extend(group_excludes('filetype', to_list(self.exclude_filetypes)))
        filters.extend([f'-{w}' for w in to_list(self.exclude_all_keywords)])

        filters.extend([f'-inurl:{w}' for w in to_list(self.not_in_url)])
        filters.extend([f'-intitle:{w}' for w in to_list(self.not_in_title)])
        filters.extend([f'-intext:{w}' for w in to_list(self.not_in_text)])

        return ' '.join([f for f in filters if f])



from datetime import date

import pytest

from search_ai import Filters


@pytest.mark.parametrize(
    'field, value, expected',
    [
        ('sites', 'example_a.com', 'site:example_a.com'),
        ('sites', ['example_a.com', 'example_b.com'], '(site:example_a.com | site:example_b.com)'),
        ('tlds', '.gov', 'site:.gov'),
        ('tlds', ['.gov', '.edu'], '(site:.gov | site:.edu)'),
        ('filetype', 'pdf', 'filetype:pdf'),
        ('https_only', True, 'inurl:https'),
        ('all_keywords', 'ai', '"ai"'),
        ('all_keywords', ['ai', 'ml'], '"ai" "ml"'),
        ('all_keywords', ['ai'], '"ai"'),
        ('any_keywords', 'ai', '"ai"'),
        ('any_keywords', ['ai', 'ml'], '("ai" | "ml")'),
        ('any_keywords', ['ai'], '"ai"'),
        ('exact_phrases', 'openai api', '"openai api"'),
        ('exact_phrases', ['foo bar', 'baz qux'], '"foo bar" "baz qux"'),
        ('in_title', 'research', 'intitle:research'),
        ('in_title', ['ai', 'ml'], 'intitle:ai intitle:ml'),
        ('in_url', 'docs', 'inurl:docs'),
        ('in_url', ['api', 'ref'], 'inurl:api inurl:ref'),
        ('in_text', 'hello', 'intext:hello'),
        ('in_text', ['world', 'vector'], 'intext:world intext:vector'),
        ('exclude_sites', 'spam.com', '-site:spam.com'),
        ('exclude_sites', ['a.com', 'b.com'], '-site:a.com -site:b.com'),
        ('exclude_tlds', '.biz', '-site:.biz'),
        ('exclude_tlds', ['.xyz', '.info'], '-site:.xyz -site:.info'),
        ('exclude_filetypes', 'exe', '-filetype:exe'),
        ('exclude_filetypes', ['bin', 'dat'], '-filetype:bin -filetype:dat'),
        ('exclude_https', True, '-inurl:https'),
        ('exclude_all_keywords', 'ads', '-ads'),
        ('exclude_all_keywords', ['spam', 'click'], '-spam -click'),
        ('exclude_all_keywords', ['ads'], '-ads'),
        ('exclude_exact_phrases', 'bad ad', '-"bad ad"'),
        ('exclude_exact_phrases', ['fake news', 'scam'], '-"fake news" -"scam"'),
        ('not_in_title', 'promo', '-intitle:promo'),
        ('not_in_title', ['clickbait', 'ad'], '-intitle:clickbait -intitle:ad'),
        ('not_in_url', 'track', '-inurl:track'),
        ('not_in_url', ['ref', 'share'], '-inurl:ref -inurl:share'),
        ('not_in_text', 'cookie', '-intext:cookie'),
        ('not_in_text', ['ads', 'popup'], '-intext:ads -intext:popup'),
    ],
)
def test_individual_fields(field, value, expected):
    filter_obj = Filters(**{field: value})
    compiled_filters = filter_obj.compile_filters()
    assert compiled_filters == expected


@pytest.mark.parametrize(
    'field, value',
    [
        ('filetype', 'toolongfilename123'),
        ('filetype', 'we!rd'),
        ('exclude_filetypes', 'we!rd'),
        ('exclude_filetypes', ['bad!', '!!']),
        ('all_keywords', 'with space'),
        ('all_keywords', 'with space'),
        ('any_keywords', ['ok', 'bad word']),
        ('any_keywords', ['ok', 'bad word']),
        ('exclude_all_keywords', 'white space'),
        ('exclude_all_keywords', ['fine', 'break this']),
        ('tlds', '.invalidtld'),
        ('tlds', ['.edu', '.badzone']),
        ('exclude_tlds', '.invalidtld'),
        ('exclude_tlds', ['.edu', '.badzone']),
    ],
)
def test_field_validation(field, value):
    with pytest.raises(ValueError, match=r'.*'):
        Filters(**{field: value})


@pytest.mark.parametrize(
    'kwargs, expected',
    [
        ({'sites': 'example.com', 'all_keywords': 'ai'}, 'site:example.com "ai"'),
        ({'tlds': '.com', 'all_keywords': ['ai', 'llm']}, 'site:.com "ai" "llm"'),
        ({'tlds': '.com', 'any_keywords': ['ai', 'llm']}, 'site:.com ("ai" | "llm")'),
        (
            {
                'sites': ['a.com', 'b.com'],
                'filetype': 'pdf',
                'exact_phrases': ['openai api', 'machine learning'],
                'https_only': True,
                'in_title': ['research', 'ai'],
            },
            '(site:a.com | site:b.com) filetype:pdf "openai api" "machine learning" inurl:https intitle:research intitle:ai',
        ),
        (
            {
                'exclude_sites': ['spam.com'],
                'exclude_all_keywords': ['ads', 'clickbait'],
                'not_in_url': 'ref',
                'exclude_https': True,
            },
            '-site:spam.com -inurl:https -ads -clickbait -inurl:ref',
        ),
        (
            {
                'filetype': 'ppt',
                'exclude_filetypes': ['exe', 'bat'],
                'any_keywords': ['presentation', 'slides'],
            },
            'filetype:ppt ("presentation" | "slides") -filetype:exe -filetype:bat',
        ),
        (
            {'in_url': 'docs', 'in_title': 'introduction', 'exclude_exact_phrases': 'outdated info'},
            'intitle:introduction inurl:docs -"outdated info"',
        ),
        (
            {'sites': 'example.org', 'in_text': 'summary', 'exclude_all_keywords': 'draft'},
            'site:example.org intext:summary -draft',
        ),
        (
            {'sites': 'example.org', 'in_text': 'summary', 'exclude_all_keywords': ['draft', 'ai', 'bot']},
            'site:example.org intext:summary -draft -ai -bot',
        ),
        (
            {
                'sites': ['news.com', 'media.com'],
                'in_url': ['breaking', 'live'],
                'not_in_text': ['subscribe', 'cookie'],
            },
            '(site:news.com | site:media.com) inurl:breaking inurl:live -intext:subscribe -intext:cookie',
        ),
        (
            {
                'exact_phrases': 'deep learning',
                'exclude_exact_phrases': 'deprecated method',
                'in_title': 'tutorial',
                'not_in_title': 'sponsored',
            },
            '"deep learning" intitle:tutorial -"deprecated method" -intitle:sponsored',
        ),
    ],
)
def test_combined_fields(kwargs, expected):
    filter_obj = Filters(**kwargs)
    compiled_filters = filter_obj.compile_filters()
    assert compiled_filters == expected



import asyncio
import subprocess
from typing import overload

from .proxy import Proxy
from .utils import valid_type

from playwright.async_api import async_playwright, Browser as AsyncBrowser, Error as PlaywrightError


PLAYWRIGHT_CONFIG = {
    'user_agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36',
    'viewport': {'width': 1280, 'height': 720},
    'locale': 'en-US',
}


@overload
def get_page_sync(url: str, proxy: Proxy | None) -> str: ...
@overload
def get_page_sync(url: list[str], proxy: Proxy | None) -> list[str]: ...
@overload
async def get_page(url: str, proxy: Proxy | None) -> str: ...
@overload
async def get_page(url: list[str], proxy: Proxy | None) -> list[str]: ...


def get_page_sync(url: str | list[str], proxy: Proxy | None) -> str | list[str]:
    try:
        loop = asyncio.get_event_loop()
        if loop.is_running():
            import nest_asyncio

            nest_asyncio.apply()
            return asyncio.run(get_page(url, proxy))
        else:
            return loop.run_until_complete(get_page(url, proxy))
    except RuntimeError:
        return asyncio.run(get_page(url, proxy))


async def get_page(url: str | list[str], proxy: Proxy | None) -> str | list[str]:
    semaphore = asyncio.Semaphore(8)
    url_list = url if isinstance(url, list) else [url]

    try:
        async with async_playwright() as playwright:
            browser = await playwright.chromium.launch(
                headless=True, proxy=proxy.to_playwright_proxy() if proxy else None
            )

            tasks = [_get_page_source(url, semaphore, browser) for url in url_list]
            results = await asyncio.gather(*tasks)
            await browser.close()

            return results[0] if isinstance(url, str) else results

    except PlaywrightError:
        print("Playwright browser not found. Running 'playwright install'...")
        subprocess.run(['playwright', 'install', 'chromium'], check=True)
        return await get_page(url, proxy)


async def _get_page_source(url: str, semaphore: asyncio.Semaphore, browser: AsyncBrowser) -> str:
    if not valid_type(url):
        return ''

    async with semaphore:
        context = await browser.new_context(**PLAYWRIGHT_CONFIG)
        page = await context.new_page()

        try:
            await page.goto(url, wait_until='domcontentloaded', timeout=6000)
            await page.wait_for_timeout(2000)  # Wait an extra 2s for JS to populate content
        except Exception:
            page_source = ''
        else:
            page_source = await page.content()

        await page.close()
        await context.close()
        return page_source



import pytest

from search_ai import Proxy


@pytest.fixture
def proxy_with_auth() -> Proxy:
    return Proxy(protocol='http', host='127.0.0.1', port=8080, username='user', password='pass')


@pytest.fixture
def proxy_no_auth() -> Proxy:
    return Proxy(protocol='https', host='proxy.io', port=443)


def test_to_httpx_proxy_url_with_auth(proxy_with_auth):
    assert proxy_with_auth.to_httpx_proxy_url() == 'http://user:pass@127.0.0.1:8080'


def test_to_httpx_proxy_url_no_auth(proxy_no_auth):
    assert proxy_no_auth.to_httpx_proxy_url() == 'https://proxy.io:443'


def test_to_playwright_proxy_with_auth(proxy_with_auth):
    expected = {'server': 'http://127.0.0.1:8080', 'username': 'user', 'password': 'pass'}
    assert proxy_with_auth.to_playwright_proxy() == expected


def test_to_playwright_proxy_no_auth(proxy_no_auth):
    expected = {'server': 'https://proxy.io:443'}
    assert proxy_no_auth.to_playwright_proxy() == expected



from lxml import html


LINK_TITLE_XPATH = '//a[@class="result-link" and not(ancestor::tr[@class="result-sponsored"])]'
DESCRIPTION_XPATH = '//td[@class="result-snippet" and not(ancestor::tr[@class="result-sponsored"])]'


def parse_search(data):
    results = []

    tree = html.fromstring(data)

    title_results = tree.xpath(LINK_TITLE_XPATH)
    desc_results = tree.xpath(DESCRIPTION_XPATH)

    assert len(title_results) == len(desc_results), 'Error parsing search results'

    for title, desc in zip(title_results, desc_results):
        results.append(
            {
                'title': title.text_content().strip(),
                'link': title.get('href'),
                'description': desc.text_content().strip(),
            }
        )

    return results



from typing import Literal, Optional
from pydantic import BaseModel, Field


class Proxy(BaseModel):
    protocol: Literal['http', 'https', 'socks4', 'socks5'] = Field(..., description='Proxy protocol')
    host: str = Field(..., description='Proxy server host or IP')
    port: int = Field(..., ge=1, le=65535, description='Proxy server port')
    username: Optional[str] = Field(None, description='Username for proxy auth')
    password: Optional[str] = Field(None, description='Password for proxy auth')

    def to_httpx_proxy_url(self) -> str:
        auth = f'{self.username}:{self.password}@' if self.username and self.password else ''
        return f'{self.protocol}://{auth}{self.host}:{self.port}'

    def to_playwright_proxy(self) -> dict:
        proxy = {'server': f'{self.protocol}://{self.host}:{self.port}'}

        if self.username:
            proxy['username'] = self.username
        if self.password:
            proxy['password'] = self.password

        return proxy



import time
import random
import asyncio
from typing import Literal, Any
from functools import partial

import curl_cffi as curl
from curl_cffi.requests.exceptions import HTTPError
from tenacity import retry, stop_after_attempt, wait_fixed, retry_if_exception_type

from .proxy import Proxy
from .filters import Filters
from .parse import parse_search
from .search_result import SearchResult, SearchResults, AsyncSearchResult, AsyncSearchResults


BASE_URL = 'https://lite.duckduckgo.com/lite/'
SLEEP_TIME = 0.25

HEADERS = {'origin': 'https://lite.duckduckgo.com', 'referer': 'https://lite.duckduckgo.com/'}

retry_curl = partial(
    retry,
    stop=stop_after_attempt(2),
    wait=wait_fixed(2),
    retry=retry_if_exception_type(HTTPError),
    reraise=True,
)


def search(
    query: str = '',
    filters: Filters | None = None,
    count: int = 10,
    offset: int = 0,
    proxy: Proxy | None = None,
) -> SearchResults:
    results = SearchResults(results=[], _proxy=proxy)

    while len(results) < count:
        response = _request(query, filters, offset, proxy)
        new_results = parse_search(response)

        if not new_results:
            return results

        for new_result in new_results:
            results.append(SearchResult(**new_result, _proxy=proxy))
            if len(results) == count:
                return results

        offset += len(new_results)

        twenty_percent = SLEEP_TIME * 0.10
        time.sleep(random.uniform(SLEEP_TIME - twenty_percent, SLEEP_TIME + twenty_percent))

    return results


async def async_search(
    query: str = '',
    filters: Filters | None = None,
    count: int = 10,
    offset: int = 0,
    proxy: Proxy | None = None,
) -> AsyncSearchResults:
    results = AsyncSearchResults(results=[], _proxy=proxy)

    while len(results) < count:
        response = await _async_request(query, filters, offset, proxy)
        new_results = parse_search(response)

        if not new_results:
            return results

        for new_result in new_results:
            results.append(AsyncSearchResult(**new_result, _proxy=proxy))
            if len(results) == count:
                return results

        offset += len(new_results)

        twenty_percent = SLEEP_TIME * 0.10
        await asyncio.sleep(random.uniform(SLEEP_TIME - twenty_percent, SLEEP_TIME + twenty_percent))

    return results


@retry_curl()
def _request(
    query: str,
    filters: Filters | None,
    offset: int,
    proxy: Proxy | None,
) -> str:
    data: dict[str, Any] = {'q': query + (filters.compile_filters() if filters else '')}

    if offset:
        data['s'] = offset - 1
        data['dc'] = offset

    if filters:
        if filters.time_span:
            data['df'] = filters.time_span.value
        if filters.region:
            data['kl'] = filters.region.value

    with curl.Session(proxy=proxy.to_httpx_proxy_url() if proxy else None) as session:
        resp = session.post(BASE_URL, data=data, headers=HEADERS)
        resp.raise_for_status()
        return resp.text


@retry_curl()
async def _async_request(
    query: str,
    filters: Filters | None,
    offset: int,
    proxy: Proxy | None,
) -> str:
    data: dict[str, Any] = {'q': query + (filters.compile_filters() if filters else '')}

    if offset:
        data['s'] = offset - 1
        data['dc'] = offset

    if filters:
        if filters.time_span:
            data['df'] = filters.time_span.value
        if filters.region:
            data['kl'] = filters.region.value

    async with curl.AsyncSession(proxy=proxy.to_httpx_proxy_url() if proxy else None) as session:
        resp = await session.post(BASE_URL, data=data, headers=HEADERS)
        resp.raise_for_status()
        return resp.text



from typing import Any

from .proxy import Proxy
from .utils import extract_metadata, generate_markdown, valid_type
from .extractor import get_page_sync, get_page

from pydantic import BaseModel, HttpUrl


class BaseSearchResult(BaseModel):
    title: str
    link: HttpUrl
    description: str | None = None
    _proxy: Proxy | None = None

    def __str__(self):
        """
        This is here to hide _proxy, which is more of an implementation detail.
        """
        return f'{self.__class__.__name__}(title="{self.title}", link="{self.link}", description="{self.description}")'

    def __repr__(self):
        return self.__str__()

    def _basic_markdown(self) -> str:
        parts = [f'**Title:** {self.title}', f'**Link:** {self.link}']
        if self.description:
            parts.append(f'**Description:** {self.description}')
        return '\n'.join(parts)

    def _extended_markdown(
        self,
        page_source: str,
        content_length: int,
        only_page_content: bool = False,
        ignore_links: bool = False,
        ignore_images: bool = True,
    ) -> str:
        markdown = generate_markdown(page_source, ignore_links, ignore_images)

        if only_page_content:
            return markdown[:content_length]

        metadata = extract_metadata(page_source)

        parts = [f'**Title:** {metadata.get("title") or self.title}', f'**Link:** {self.link}']

        if metadata.get('description'):
            parts.append(f'**Description:** {metadata["description"]}')
        elif self.description:
            parts.append(f'**Description:** {self.description}')

        if metadata.get('author'):
            parts.append(f'**Author:** {metadata["author"]}')
        if metadata.get('twitter'):
            parts.append(f'**Twitter:** {metadata["twitter"]}')

        if markdown:
            parts.append('')  # Extra break between metadata and page data
            parts.append('## Page Preview:\n')
            parts.append(markdown[:content_length].strip())

        return '\n'.join(parts)

    def _extended_json(
        self, page_source: str, content_length: int, ignore_links: bool = False, ignore_images: bool = True
    ) -> dict:
        metadata = extract_metadata(page_source)
        markdown = generate_markdown(page_source, ignore_links, ignore_images)

        combined_data = {'title': metadata.get('title') or self.title, 'link': str(self.link)}

        if metadata.get('description'):
            combined_data['description'] = metadata['description']
        elif self.description:
            combined_data['description'] = self.description

        if metadata.get('author'):
            combined_data['author'] = metadata['author']
        if metadata.get('twitter'):
            combined_data['twitter'] = metadata['twitter']

        if markdown:
            combined_data['page_preview'] = markdown[:content_length]

        return combined_data


class SearchResult(BaseSearchResult):
    def markdown(
        self,
        extend: bool = False,
        content_length: int = 1_000,
        ignore_links: bool = False,
        ignore_images: bool = True,
        only_page_content: bool = False,
    ) -> str:
        if not extend or not valid_type(str(self.link)):
            return self._basic_markdown()

        page_source = get_page_sync(str(self.link), self._proxy)
        return self._extended_markdown(page_source, content_length, only_page_content, ignore_links, ignore_images)

    def json(
        self,
        extend: bool = False,
        content_length: int = 1_000,
        ignore_links: bool = False,
        ignore_images: bool = True,
        **kwargs: Any,
    ) -> dict:
        if not extend or not valid_type(str(self.link)):
            return super().model_dump(**kwargs)

        page_source = get_page_sync(str(self.link), self._proxy)
        return self._extended_json(page_source, content_length, ignore_links, ignore_images)


class AsyncSearchResult(BaseSearchResult):
    async def markdown(
        self,
        extend: bool = False,
        content_length: int = 1_000,
        ignore_links: bool = False,
        ignore_images: bool = True,
        only_page_content: bool = False,
    ) -> str:
        if not extend or not valid_type(str(self.link)):
            return self._basic_markdown()

        page_source = await get_page(str(self.link), self._proxy)
        return self._extended_markdown(page_source, content_length, only_page_content, ignore_links, ignore_images)

    async def json(
        self,
        extend: bool = False,
        content_length: int = 1_000,
        ignore_links: bool = False,
        ignore_images: bool = True,
        **kwargs: Any,
    ) -> dict:
        if not extend or not valid_type(str(self.link)):
            return super().model_dump(**kwargs)

        page_source = await get_page(str(self.link), self._proxy)
        return self._extended_json(page_source, content_length, ignore_links, ignore_images)


class SearchResults(list):
    def __init__(self, results: list[SearchResult], _proxy: Proxy | None = None):
        super().__init__(results)
        self._proxy = _proxy

    def markdown(self, extend: bool = False, content_length: int = 1_000, **kwargs) -> str:
        if not self:  # Edge case for no search results
            return ''

        if not extend:
            content = [result._basic_markdown() for result in self]

        else:
            page_sources = get_page_sync([str(result.link) for result in self], self._proxy)
            content = [
                result._extended_markdown(page_source=page_source, content_length=content_length, **kwargs)
                for result, page_source in zip(self, page_sources)
            ]

        return '# Search Results\n\n' + '\n----------\n'.join(content)

    def json(self, extend: bool = False, content_length: int = 1_000, **kwargs) -> list[dict]:
        if not extend:
            return [result.model_dump() for result in self]

        page_sources = get_page_sync([str(result.link) for result in self], self._proxy)
        return [
            result._extended_json(page_source=page_source, content_length=content_length, **kwargs)
            for result, page_source in zip(self, page_sources)
        ]


class AsyncSearchResults(list):
    def __init__(self, results: list[AsyncSearchResult], _proxy: Proxy | None = None):
        super().__init__(results)
        self._proxy = _proxy

    async def markdown(self, extend: bool = False, content_length: int = 1_000, **kwargs) -> str:
        if not extend:
            content = [result._basic_markdown() for result in self]

        else:
            page_sources = await get_page([str(result.link) for result in self], self._proxy)
            content = [
                result._extended_markdown(page_source=page_source, content_length=content_length, **kwargs)
                for result, page_source in zip(self, page_sources)
            ]

        return '# Search Results\n\n' + '\n----------\n'.join(content)

    async def json(self, extend: bool = False, content_length: int = 1_000, **kwargs) -> list[dict]:
        if not extend:
            return [result.model_dump() for result in self]

        page_sources = await get_page([str(result.link) for result in self], self._proxy)
        return [
            result._extended_json(page_source=page_source, content_length=content_length, **kwargs)
            for result, page_source in zip(self, page_sources)
        ]



import pytest
from pydantic import HttpUrl
from search_ai.search_result import (
    SearchResult,
    AsyncSearchResult,
    SearchResults,
    AsyncSearchResults,
)


@pytest.fixture
def sample_data():
    return {'title': 'Example Title', 'link': 'https://example.com', 'description': 'An example description.'}


def test_search_result(sample_data):
    result = SearchResult(**sample_data)

    assert result.title == sample_data['title']
    assert result.link == HttpUrl(sample_data['link'])
    assert result.description == sample_data['description']

    markdown = result.markdown()
    assert '**Title:**' in markdown
    assert '**Link:**' in markdown
    assert '**Description:**' in markdown

    json_data = result.json()
    for key, val in sample_data.items():
        if key == 'link':
            assert json_data[key] == HttpUrl(val)
        else:
            assert json_data[key] == val

    assert set(json_data.keys()) == set(sample_data.keys())


@pytest.mark.asyncio
async def test_async_search_result(sample_data):
    result = AsyncSearchResult(**sample_data)
    assert result.title == sample_data['title']
    assert result.link == HttpUrl(sample_data['link'])
    assert result.description == sample_data['description']

    markdown = await result.markdown()
    assert '**Title:**' in markdown
    assert '**Link:**' in markdown
    assert '**Description:**' in markdown

    json_data = await result.json()
    for key, val in sample_data.items():
        if key == 'link':
            assert json_data[key] == HttpUrl(val)
        else:
            assert json_data[key] == val

    assert set(json_data.keys()) == set(sample_data.keys())


def test_search_results(sample_data):
    r1 = SearchResult(**sample_data)
    r2 = SearchResult(**sample_data)
    results = SearchResults(results=[r1, r2])

    markdown = results.markdown()
    assert markdown.count('**Title:**') == 2
    assert markdown.startswith('# Search Results')

    json_data = results.json()
    assert isinstance(json_data, list)
    assert len(json_data) == 2


@pytest.mark.asyncio
async def test_async_search_results(sample_data):
    r1 = AsyncSearchResult(**sample_data)
    r2 = AsyncSearchResult(**sample_data)
    results = AsyncSearchResults(results=[r1, r2])

    markdown = await results.markdown()
    assert markdown.count('**Title:**') == 2
    assert markdown.startswith('# Search Results')

    json_data = await results.json()
    assert isinstance(json_data, list)
    assert len(json_data) == 2



import random
import mimetypes

import html2text
from lxml import html, etree


def extract_metadata(page_source: str) -> dict:
    try:
        tree = html.fromstring(page_source)
    except Exception:
        return {}

    page_title = tree.xpath('//head/title/text()')
    page_description = tree.xpath('//head/meta[@name="description"]/@content')
    author = tree.xpath('//head/meta[@name="author"]/@content')
    twitter_handle = tree.xpath('//head/meta[@name="twitter:site"]/@content')

    result = {}

    if page_title:
        result['title'] = page_title[0]
    if page_description and valid_description_metadata(page_description[0]):
        result['description'] = page_description[0]
    if author:
        result['author'] = author[0]
    if twitter_handle:
        result['twitter'] = twitter_handle[0]

    return result


def valid_description_metadata(desc: str) -> bool:
    try:
        etree.fromstring(desc)
        return False
    except etree.XMLSyntaxError:
        return True


def generate_markdown(page_source: str, ignore_links: bool, ignore_images: bool) -> str:
    text_maker = html2text.HTML2Text()
    text_maker.ignore_links = ignore_links
    text_maker.ignore_images = ignore_images
    text_maker.body_width = 0  # Prevent automatic wrapping
    return text_maker.handle(page_source).strip()


def valid_type(url: str) -> bool:
    """
    Currently, only text / html pages are supported for extended data retrival.
    """
    mime, _ = mimetypes.guess_type(url)
    return mime in ('text/html', 'text/plain', 'application/xhtml+xml', None)
